unit EntityManager;

interface

uses
	SysUtils, Classes, fgl, Math, Math2D, dialogs;

type
	TEntity = class(TPersistent)
	private
		fPos, fTarget: TPointF;
		fAngle, fRange, fSpeed, fXspeed, fYSpeed: Double;
		fActive: Boolean;
	public
                constructor Create(APos, ATarget: TPointF; AAngle, ARange, ASpeed, AxSpeed,
			AySpeed: Double; AActive: Boolean);
		property Pos: TPointF read fPos write fPos;
		property Target: TPointF read fTarget write fTarget;
		property Angle: Double read fAngle write fAngle;
		property Range: Double read fRange write fRange;
		property Speed: Double read fSpeed write fSpeed;
		property xSpeed: Double read fXspeed write fXspeed;
		property ySpeed: Double read fYSpeed write fYSpeed;
		property Active: Boolean read fActive write fActive;
	end;

        TEntities = specialize TFPGObjectList<TEntity>;

	TEntityManager = class(TPersistent)
	private
		fEntities: TEntities;
		procedure Dispose;
	public
		constructor Create;
		destructor Destroy; override;
	published
		procedure Add(APosition, ATarget: TPointF; AAngle, ARange, ASpeed: Double);
		procedure Update;
		property Entities: TEntities read fEntities write fEntities;
	end;

implementation

constructor TEntity.Create(APos, ATarget: TPointF;
	AAngle, ARange, ASpeed, AxSpeed, AySpeed: Double; AActive: Boolean);
begin
	inherited Create;
	fPos := APos;
	fTarget := ATarget;
	fAngle := AAngle;
	fRange := ARange;
	fSpeed := ASpeed;
	fXspeed := AxSpeed;
	fYSpeed := AySpeed;
	fActive := AActive;
end;

procedure TEntityManager.Dispose;
var
	I: Integer;
begin
	for I := fEntities.Count - 1 downto 0 do
		if fEntities[I].Active = False then
		begin
			fEntities.Delete(I);
		end;
end;

constructor TEntityManager.Create;
begin
	inherited Create;
	fEntities := TEntities.Create(True);
end;

destructor TEntityManager.Destroy;
begin
	FreeAndNil(fEntities);
	inherited Destroy;
end;

procedure TEntityManager.Add(APosition, ATarget: TPointF;
	AAngle, ARange, ASpeed: Double);
var
	D: Double;
	sinTheta, cosTheta: Extended;
begin

	D := DistancePointsF(APosition.X, APosition.Y, ATarget.X, ATarget.Y);
	if D > ARange then
		ATarget := PointOfAngle(APosition.X, APosition.Y, ARange, AAngle)
	else
		ATarget := ATarget;

	SinCos(DegToRad(AAngle), sinTheta, cosTheta);
	fEntities.Add(TEntity.Create(APosition, ATarget, AAngle, ARange, ASpeed,
		cosTheta * ASpeed, sinTheta * ASpeed, True))
end;

procedure TEntityManager.Update;
var
	I: Integer;
	xc, yc: Double;
begin
	for I := 0 to fEntities.Count - 1 do
	begin
		with fEntities[I] do
		begin
			if fEntities[I].Active then
			begin

				if PointsEqualF(Pos, Target) then
				begin
					Active := False;
				end
				else
				begin
					if fEntities[I].Pos.X <> fEntities[I].Target.X then
					begin
						if (xSpeed > 0) and (xSpeed >= Target.X - Pos.X) then
							xc := Pos.X + (Target.X - Pos.X)
						else if (xSpeed < 0) and (xSpeed <= Target.X - Pos.X) then
							xc := Pos.X + (Target.X - Pos.X)
						else
							xc := Pos.X + xSpeed;
						Pos := PointF(xc, Pos.Y);
					end;

					if fEntities[I].Pos.Y <> fEntities[I].Target.Y then
					begin
						if (ySpeed > 0) and (ySpeed >= Target.Y - Pos.Y) then
							yc := Pos.Y + (Target.Y - Pos.Y)
						else if (ySpeed < 0) and (ySpeed <= Target.Y - Pos.Y) then
							yc := Pos.Y + (Target.Y - Pos.Y)
						else
							yc := Pos.Y + ySpeed;
						Pos := PointF(Pos.X, yc);
					end;
				end;
			end;
		end;
	end;
	Dispose;
end;

end.
